/*
   huffman.c
   Implements the huffman algorithm for gsLib
*/
#include <stdlib.h>
#include <string.h>
#include <compress.h>

#define MAX_BITS 15
static int max_length = MAX_BITS;

/*
   static int HuffSortByLength(const void *a,const void *b):
   The sorter function for use with the qsort function. Sorts a HuffTable by
   length
*/
static int HuffSortByLength(const void *a,const void *b)
{
   return (((HuffTable *)a)->len < ((HuffTable *)b)->len ? -1 :
            (((HuffTable *)a)->len > ((HuffTable *)b)->len ? 1 : 0));
}

/*
   static int HuffSortByCode(const void *a,const void *b):
   The sorter function for use with the qsort function. Sorts a HuffTable by
   code
*/
static int HuffSortByCode(const void *a,const void *b)
{
   return (((HuffTable *)a)->code < ((HuffTable *)b)->code ? -1 :
            (((HuffTable *)a)->code > ((HuffTable *)b)->code ? 1 : 0));
}

/*
   static dword ReverseBits(dword code,int len):
   Reverses the bits of `len' length of the code
*/
static dword ReverseBits(dword code,int len)
{
   register dword bit = 0;

   do {
      bit |= code & 1;
      code >>= 1;
      bit <<= 1;
   } while(--len);

   return bit >> 1;
}

/*
   void CalculateMinimumRedundancy(HuffTable *tab,int n):
   Calculated the minimum bits required to represent each code
   See http://www.cs.mu.oz.au/~alistair/abstracts/wads95.html
*/
void CalculateMinimumRedundancy(HuffTable *tab,int n)
{
	int root, leaf, next, avbl, used, dpth;
	HuffTable *hsave;
	int nsave;

	if(n == 0) return;
	if(n == 1) {
		tab[0].len = 1;
		return;
	}

	qsort(tab,n,sizeof(HuffTable),HuffSortByLength);
	for(used = 0;used < n;used++) if(tab[used].len) break;
	if(used >= n-1) {
		tab[n-1].len = 1;
		qsort(tab,n,sizeof(HuffTable),HuffSortByCode);
		return;
	}

	hsave = tab;
	nsave = n;
	for(;!(*tab).len;tab++,n--);

	tab[0].len += tab[1].len;
	root = 0;
	leaf = 2;

	for(next = 1; next < n-1; next++) {
		if(leaf >= n || tab[root].len < tab[leaf].len) {
			tab[next].len = tab[root].len;
			tab[root++].len = next;
		} else {
			tab[next].len = tab[leaf++].len;
		}

		if(leaf >= n || (root < next && tab[root].len < tab[leaf].len)) {
			tab[next].len += tab[root].len;
			tab[root++].len = next;
		} else {
			tab[next].len += tab[leaf++].len;
		}
	}

	tab[n-2].len = 0;
	for(next = n-3; next >= 0; next--)
		tab[next].len = tab[tab[next].len].len + 1;

	avbl = 1;
	used = dpth = 0;
	root = n-2;
	next = n-1;
	while(avbl > 0) {
		while(root >= 0 && tab[root].len == dpth) {
			used++;
			root--;
		}
		while(avbl > used) {
			tab[next--].len = dpth;
			avbl--;
		}
		avbl = 2*used;
		dpth++;
		used = 0;
	}

	tab = hsave;
	n = nsave;

	qsort(tab,n,sizeof(HuffTable),HuffSortByCode);
}

/*
   void BuildHuffTable(HuffTable *tab,int n):
   Builds huffman codes with respect to their length
   Based on algorithm in the RFC-Deflate document
*/
void BuildHuffTable(HuffTable *tab,int n)
{
   dword code = 0;
	int bits, i, len, overflow = 0;
	int bl_count[MAX_BITS + 1] = {0};
	int next_code[MAX_BITS + 1] = {0};

	// Build length table
	for(i = 0;i < n;i++) {
		if(tab[i].len > max_length) {
			overflow++;
			continue;
		}
		bl_count[tab[i].len]++;
	}

	while(overflow > 0) {
		bits = max_length - 1;
		while(!bl_count[bits] && bits >= 0) bits--;
		for(i = 0;i < n;i++)
			if(tab[i].len == bits) { tab[i].len++; break; }
		for(i = 0;i < n;i++)
			if(tab[i].len > max_length) { tab[i].len = bits+1; break; }
		bl_count[bits]--;
		bl_count[bits+1] += 2;
		overflow--;
	}

	// calculate minimum code values
	bl_count[0] = 0;
	for(bits = 1; bits <= MAX_BITS; bits++) {
		code = (code + bl_count[bits - 1]) << 1;
		next_code[bits] = code;
	}

	for(i = 0; i < n; i++) {
		len = tab[i].len;
		if(len != 0) {
			tab[i].bits = next_code[len];
			next_code[len]++;
		}
	}
}

/*
   HuffLookup *BuildHuffmanLookupTable(HuffTable *tab,int n):
   Generates a lookup table for the FetchHuffmanCode() function
   `tab' is the pointer to a HuffTable struct of huffman codes
*/
HuffLookup *BuildHuffmanLookupTable(HuffTable *tab,int n)
{
   register HuffLookup *lookup; // the lookup table
   int i, j, pos, nexttable = 0, entry;
   dword curcode;
   int curlen;

   lookup = (HuffLookup *)malloc(sizeof(HuffLookup));
   if(!lookup) {
      gerror("Not enough memory for HuffLookup table.");
      THROW (E_MEMORY) return NULL;
   }

   memset(lookup,0,sizeof(HuffLookup));

   for(i = 0; i < n; i++) {
      if(!tab[i].len) continue;

      // if is less than 9-bits
      if(tab[i].len <= 9) {
         // shift to left align
         pos = tab[i].bits << (9 - tab[i].len);
      } else {
         // shift for right alignment
         pos = tab[i].bits >> (tab[i].len - 9);
      }

      pos &= 0x1FF;
      lookup->tablelo[pos].code = tab[i].code; // assign code
      if(tab[i].len > 9) { // if len is > 9, use hi table
         if(lookup->tablelo[pos].len & 0x8000) { // if pos has 0x8000 flaged on
            entry = lookup->tablelo[pos].len & 0x7FFF;
         } else { // otherwise allocate new entry
            lookup->tablehi[nexttable] = malloc(256);
            if(!lookup->tablehi[nexttable]) {
               FreeHuffmanLookupTable(lookup);
               gerror("Not enough memory for lookup table.");
               THROW (E_MEMORY) return NULL;
            }
            memset(lookup->tablehi[nexttable],0,256);
            lookup->tablelo[pos].len = 0x8000 | nexttable; // assign len
            entry = nexttable;
            nexttable++;
         }

         // assign code and length
         lookup->tablehi[entry]
               [(tab[i].bits & ((1 << (tab[i].len - 9)) - 1)) << (15 - tab[i].len)].code = tab[i].code;
         lookup->tablehi[entry]
               [(tab[i].bits & ((1 << (tab[i].len - 9)) - 1)) << (15 - tab[i].len)].len = tab[i].len;
      } else {
         lookup->tablelo[pos].len = tab[i].len; // assign len
      }
   }

   curcode = 0;
   curlen = 0;
   // we duplicate entries throughout the table
   for(i = 0; i < 512; i++) {
      if(lookup->tablelo[i].len) {
         if(lookup->tablelo[i].code != curcode || lookup->tablelo[i].len != curlen) {
            curcode = lookup->tablelo[i].code;
            curlen = lookup->tablelo[i].len;
         }
      }
      lookup->tablelo[i].code = curcode;
      lookup->tablelo[i].len = curlen;
   }

   curcode = 0;
   curlen = 0;
   // duplicate for the hi table
   for(i = 0; i < nexttable; i++) {
      for(j = 0; j < 64; j++) {
         if(lookup->tablehi[i][j].len) {
            if(lookup->tablehi[i][j].code != curcode || lookup->tablehi[i][j].len != curlen) {
               curcode = lookup->tablehi[i][j].code;
               curlen = lookup->tablehi[i][j].len;
            }
         }
         lookup->tablehi[i][j].code = curcode;
         lookup->tablehi[i][j].len = curlen;
      }
   }

   return lookup;
}

/*
   void FreeHuffmanLookupTable(HuffLookup *lookup):
   Frees the memory used by a lookup table
*/
void FreeHuffmanLookupTable(HuffLookup *lookup)
{
   int i;

   for(i = 0; i < 256 && lookup->tablehi[i]; i++) free(lookup->tablehi[i]);
   free(lookup);
}

/*
   dword FetchHuffmanCode(STREAM *fp,HuffLookup *lookup):
   Gets a huffman code from the stream using the lookup table
*/
dword FetchHuffmanCode(STREAM *fp,HuffLookup *lookup)
{
   dword code;
   int bitsToReturn;
   dword value;
   int len;

   code = ReverseBits(stream_read_bitr(fp,9),9); // read 9 bits

   len = lookup->tablelo[code].len;
   if(len > 9) { // if > 9 use hi table
      code = ReverseBits(stream_read_bitr(fp,6),6); // 6 bits
      value = lookup->tablehi[len & 0x7FFF][code].code;
      len = lookup->tablehi[len & 0x7FFF][code].len;
      bitsToReturn = 15 - len;
      code = ReverseBits(code,6) >> (6 - bitsToReturn); // unreverse bits
   } else {
      value = lookup->tablelo[code].code;
      bitsToReturn = 9 - len;
      code = ReverseBits(code,9) >> len;
   }

   stream_unread_bitr(fp,code,bitsToReturn);

   return value;
}

/*
   void WriteHuffmanCode(STREAM *fp,HuffTable tab):
   Writes the huffman code to the stream
*/
void WriteHuffmanCode(STREAM *fp,HuffTable tab)
{
   dword code;

   code = ReverseBits(tab.bits & ((1 << tab.len) - 1),tab.len);
   stream_write_bitr(fp,code,tab.len);
}

/*
   static void ReadCodeLengths(STREAM *fp,HuffLookup *l,HuffTable tab[],int n):
   Reads the huffman code lengths to tab
*/
static void ReadCodeLengths(STREAM *fp,HuffLookup *l,HuffTable tab[],int n)
{
   int i, len, code, lastcode = 0;

   for(i = 0; i < n; i++) {
      code = FetchHuffmanCode(fp,l);
      switch(code) {
         default: tab[i].len = code; lastcode = code; break;
         case 16: len = stream_read_bitr(fp,2) + 3;
                  for(;len;i++,len--) tab[i].len = lastcode;
                  i--;
                  break;
         case 17: len = stream_read_bitr(fp,3) + 3;
                  for(;len;i++,len--) tab[i].len = 0;
                  i--;
                  break;
         case 18: len = stream_read_bitr(fp,7) + 11;
                  for(;len;i++,len--) tab[i].len = 0;
                  i--;
                  break;
      } // end switch
   } // end for
}

/*
   static void WriteCodeLengths(STREAM *fp,HuffTable *lentab,HuffTable *tab,int n,int scan):
   Writes the code lengths to file or builds a frequency of code lengths if
   `scan' = TRUE
*/
static void WriteCodeLengths(STREAM *fp,HuffTable *lentab,HuffTable *tab,int n,int scan)
{
   int i, runcount, current, match = 0;

   current = tab[0].len; // get first length
   for(i = 1; i <= n; i++) {
      runcount = 0;

      for( ; i <= n; i++) {
         runcount++; // increment runcount
         match = tab[i].len;

         if(runcount >= 7 && current != 0) break; // max run is 7 for non zeros
         if(runcount >= 138) break; // max zero rle is 138
         if(match != current) break;
      }

      // if zero runcount
      if(current == 0 && runcount >= 3) {
         if(runcount <= 10) { // write code 17 if < 10
            if(scan) lentab[17].len++;
            else {
               WriteHuffmanCode(fp,lentab[17]);
               stream_write_bitr(fp,runcount - 3,3);
            }
          } else {            // otherwise write code 18 for long zrle code
            if(scan) lentab[18].len++;
            else {
               WriteHuffmanCode(fp,lentab[18]);
               stream_write_bitr(fp,runcount - 11, 7);
            }
          }
      } else if(runcount > 3 && current != 0) {
         // if 3 < runcount < 7, then emit code 16
         if(scan) {
            lentab[current].len++;
            lentab[16].len++;
         } else {
            WriteHuffmanCode(fp,lentab[current]);
            WriteHuffmanCode(fp,lentab[16]);
            stream_write_bitr(fp,runcount - 4, 2);
         }
      } else {
         // emit literals
         for(; runcount; runcount--) {
            if(scan) lentab[current].len++;
            else WriteHuffmanCode(fp,lentab[current]);
         }
      }

      current = match;
   }
}

/*
   int ReadHuffmanHeader(STREAM *fp,HuffTable *tab,HuffTable **dist,int *nLit,int *nDist):
   Reads a huffman header from a stream to `tab' and updates `dist' to point to
   the start of the distance table. If `dist' is NULL, it is ignored
   The header follows accordingly to the deflate header. `dist' will point to
   NULL if the distance table is not used.
   The return values returns a set of flags as decribed below:
      if the return is -1 and error has occured
      if the return is 0, the function was successful
      if bit 1 is set, it means that it is the final block
      if bit 2 is set, it means that the block is not compressed
*/
int ReadHuffmanHeader(STREAM *fp,HuffTable *tab,HuffTable **dist,int *nLit,int *nDist)
{
   int hclen, hlit, hdist, bfinal, btype, i;

   HuffLookup *lookup;
   HuffTable codlen[19] = {
      {16, 0}, {17, 0}, {18, 0},
      { 0, 0},	{ 8, 0},	{ 7, 0},
      { 9, 0},	{ 6, 0},	{10, 0},
      { 5, 0},	{11, 0},	{ 4, 0},
      {12, 0},	{ 3, 0},	{13, 0},
      { 2, 0},	{14, 0},	{ 1, 0},
      {15, 0}
   };

   bfinal = stream_read_bitr(fp,1); // read bfinal
   btype = stream_read_bitr(fp,2);

   switch(btype) {
      case 0   :  return 2 | bfinal; // no compression
      case 1   :  // initializes the Huffman Table
                  for(i = 0; i < 286; i++) {
                     tab[i].code = i;
                     tab[i].len = 0;
                     if(i < 30) {
                        tab[286 + i].code = i;
                        tab[286 + i].len = 0;
                     }
                  }

                  if(nLit) *nLit = 286;
                  if(nDist) *nDist = 30;
                  if(dist) *dist = &tab[286];

                  for(i = 0; i <= 143; i++) tab[i].len = 8;
                  for(i = 144; i <= 255; i++) tab[i].len = 9;
                  for(i = 256; i <= 279; i++) tab[i].len = 7;
                  for(i = 280; i <= 285; i++) tab[i].len = 8;
                  if(dist) for(i = 0; i < 30; i++) tab[286 + i].len = 5;
                  BuildHuffTable(tab,286);
                  if(dist) BuildHuffTable(*dist,30);
                  return bfinal;
      case 2   :  break;             // compressed with dynamic huffman codes
      default  :  gerror("BTYPE is unknown in inflation.");
                  THROW (E_FORMAT) return -1;
   }

   hlit = stream_read_bitr(fp,5);
   hdist = stream_read_bitr(fp,5);
   hclen = stream_read_bitr(fp,4);

   if(dist) {
      if(hdist > 0) *dist = &tab[hlit + 257];
      else *dist = NULL;
   }
   if(nLit) *nLit = hlit + 257;
   if(nDist) *nDist = hdist ? hdist + 1 : 0;

   if(!hdist) hdist = -1;

   // initializes the Huffman Table
   for(i = 0; i < hlit + 257; i++) {
      tab[i].code = i;
      tab[i].len = 0;
      if(i < hdist + 1) {
         tab[hlit + 257 + i].code = i;
         tab[hlit + 257 + i].len = 0;
      }
   }

   // read in code lengths
   for(i = 0; i < hclen + 4; i++)
      codlen[i].len = stream_read_bitr(fp,3);

   // build the huffman codes
   qsort(codlen,hclen + 4,sizeof(HuffTable),HuffSortByCode);
   BuildHuffTable(codlen,hclen + 4);
   lookup = BuildHuffmanLookupTable(codlen,hclen + 4);
   if(!lookup) return -1;

   ReadCodeLengths(fp,lookup,tab,hlit + hdist + 258);
   BuildHuffTable(tab,hlit + 257);
   if(dist) {
      if(*dist) BuildHuffTable(*dist,hdist + 1);
   }

   free(lookup);

   return bfinal;
}

/*
   int WriteHuffmanHeader(STREAM *fp,HuffTable *tab,int nLit,int nDist):
   Writes a huffman header to the file. `nLit' is the number of literal values,
   while `nDist' is the number of distance codes.
   Returns 0 on success
*/
int WriteHuffmanHeader(STREAM *fp,HuffTable *tab,int nLit,int nDist)
{
   HuffTable lentab[19]; // len table
   int hlit, hdist, hclen, i, lensav;
   const byte order[19] =
   {
      16,17,18,0,8,7,9,6,10,5,11,4,12,3,13,2,14,1,15
   };

   hlit = nLit < 257 ? 257 : nLit;
   hdist = nDist == 1 ? 2 : nDist;

   // initializes lentab
   for(i = 0; i < 19; i++) {
      lentab[i].code = i;
      lentab[i].len = 0;
   }

   // pass 1 scan code length
   WriteCodeLengths(fp,lentab,tab,hlit + hdist,TRUE);

   lensav = max_length;
   max_length = 7; // readjust max_length
   CalculateMinimumRedundancy(lentab,19);
   BuildHuffTable(lentab,19);
   max_length = lensav;

   stream_write_bitr(fp,1,1); // BFINAL = 1
   stream_write_bitr(fp,2,2); // BTYPE = dynamic huffman codes

   hclen = 19;

   stream_write_bitr(fp,hlit - 257,5);
   stream_write_bitr(fp,hdist ? hdist - 1 : 0,5);
   stream_write_bitr(fp,hclen - 4,4);

   // write lens
   for(i = 0; i < 19; i++)
      stream_write_bitr(fp,lentab[order[i]].len,3);

   WriteCodeLengths(fp,lentab,tab,hlit + hdist,FALSE);

   return 0;
}
